# -*- coding: utf-8 -*-

"""
@date: 2020/7/13 下午4:11
@file: utility.py
@author: zj
@description: Common operations
"""

import xmltodict


def xml_to_dict(xml_path):
    with open(xml_path, 'rb') as f:
        xml_parse = xmltodict.parse(f)
    return xml_parse


def dict_to_xml(data_dict, file_path):
    with open(file_path, 'w', encoding='utf-8') as f:
        xml_str = xmltodict.unparse(data_dict, output=f, pretty=1)
    return xml_str


def xyxy_2_xywh(bndbox, size):
    """
    创建yolov5
    :param bndbox: [xmin, ymin, xmax, ymax]
    :param size: (width, height)
    :return: [x_center, y_center, width, height]相对格式
    """
    xmin, ymin, xmax, ymax = bndbox
    img_w, img_h = size[:2]

    x_center = (xmin + xmax) / 2
    y_center = (ymin + ymax) / 2
    width = xmax - xmin
    height = ymax - ymin

    x_center = 1.0 * x_center / img_w
    y_center = 1.0 * y_center / img_h
    width = 1.0 * width / img_w
    height = 1.0 * height / img_h

    return [x_center, y_center, width, height]



# -*- coding: utf-8 -*-

"""
@date: 2020/7/18 下午2:00
@file: defaults.py
@author: zj
@description: 
"""

import argparse


def default_argument_parser():
    """
    Create a parser with some common arguments used by users.

    Returns:
        argparse.ArgumentParser:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', "--config_file", default="", metavar="CONFIG_FILE", help="path to config file")
    parser.add_argument('-v', '--version', help='output version information', action="store_true")
    return parser



# -*- coding: utf-8 -*-

"""
@date: 2020/7/14 下午7:14
@file: parse_voc_xml.py
@author: zj
@description: Parsing Pascal VOC format XML file
"""

import os
import collections
import xml.etree.cElementTree as ET


class ParseVocXml(object):
    """
    Parsing a single xml file
    """

    def __init__(self, xml_path):
        self.target = self.preprocess(xml_path)

    def preprocess(self, xml_path):
        if not os.path.exists(xml_path):
            raise ValueError('{} does not exist'.format(xml_path))
        if not self.check_xml_suffix(xml_path):
            raise ValueError('{} is not an XML file'.format(xml_path))

        return self.parse_voc_xml(ET.parse(xml_path).getroot())

    def check_xml_suffix(self, xml_path):
        """
        Check if it is an XML file
        """
        suffix = os.path.splitext(os.path.split(xml_path)[1])[1]
        return True if suffix == '.xml' else False

    def parse_voc_xml(self, node):
        voc_dict = {}
        children = list(node)
        if children:
            def_dic = collections.defaultdict(list)
            for dc in map(self.parse_voc_xml, children):
                for ind, v in dc.items():
                    def_dic[ind].append(v)
            if node.tag == 'annotation':
                def_dic['object'] = [def_dic['object']]
            voc_dict = {
                node.tag:
                    {ind: v[0] if len(v) == 1 else v
                     for ind, v in def_dic.items()}
            }
        if node.text:
            text = node.text.strip()
            if not children:
                voc_dict[node.tag] = text
        return voc_dict

    def get_size(self):
        """
        :return: Image width and height
        """
        size = self.target['annotation']['size']
        width = int(size['width'])
        height = int(size['height'])
        depth = int(size['depth'])
        return (width, height, depth)

    def get_objects(self, difficult=False):
        """
        Returns all dimension objects
        @:param difficult: False - Do not return bounding boxes marked as difficult
        :return: [{'name': '', 'bndbox': [xmin, ymin, xmax, ymax]}]
        """
        object_list = self.target['annotation']['object']

        res_object_list = list()
        if object_list is not None:
            if not isinstance(object_list, list):
                # Contains only a single target
                object_list = [object_list]
            for object_dict in object_list:
                isdifficult = object_dict['difficult']
                if difficult == False and isdifficult == 1:
                    continue
                name = object_dict['name']
                bndbox = self.get_bndbox(object_dict['bndbox'])
                res_object_list.append({'name': name, 'bndbox': bndbox})
        return res_object_list

    def get_bndbox(self, bndbox):
        """
        :return: Returns the coordinates of the upper left and lower right corners
        """
        xmin = int(bndbox['xmin'])
        ymin = int(bndbox['ymin'])
        xmax = int(bndbox['xmax'])
        ymax = int(bndbox['ymax'])

        if xmin > xmax:
            tmp = xmax
            xmax = xmin
            xmin = tmp
        if ymin > ymax:
            tmp = ymax
            ymax = ymin
            ymin = tmp
        return [xmin, ymin, xmax, ymax]



# -*- coding: utf-8 -*-

"""
@date: 2020/7/17 下午7:29
@file: tlt_anno.py
@author: zj
@description: 
"""

import os
import cv2
import numpy as np
import json

from . import registry
from ..anno.base_anno import BaseAnno
from ..util.misc import get_img_name, check_input_output_folder
from ..util.logger import setup_logger


@registry.ANNOS.register('tlt')
class TltAnno(BaseAnno):
    """
    [Nvidia TLT](https://docs.nvidia.com/metropolis/TLT/tlt-getting-started-guide/index.html)训练需要Kitti格式
    refer to:
    [Object Detection Data Extension](https://github.com/NVIDIA/DIGITS/blob/v4.0.0-rc.3/digits/extensions/data/objectDetection/README.md)
    [Object Detection: About KITTI format #992](https://github.com/NVIDIA/DIGITS/issues/992)
    Each line represents a annotation object, with 15 fields (separated by spaces). The first field represents the class name,
    the fifth to ninth represents xmin/Ymin/xmax/ymax (floating-point representation), and the rest is set to 0.
    Overall preservation architecture as follows:
    root/
        images/
            1.png
            2.png
        labels/
            1.txt
            2.txt
    """

    def __init__(self, cfg):
        self.name = cfg.TLT.NAME
        self.img_extension = cfg.TLT.IMG_EXTENSION
        self.anno_extension = cfg.TLT.ANNO_EXTENSION

        if cfg.ANNO.PARSER == self.name:
            self.src_dir = cfg.INPUT.DIR
            self.image_folder = cfg.INPUT.IMAGE_FOLDER
            self.label_folder = cfg.INPUT.LABEL_FOLDER
        if cfg.ANNO.CREATOR == self.name:
            self.dst_dir = cfg.OUTPUT.DIR
            self.image_folder = cfg.OUTPUT.IMAGE_FOLDER
            self.label_folder = cfg.OUTPUT.LABEL_FOLDER

        self.verbose = cfg.ANNO.VERBOSE

        self.logger = setup_logger(__name__)
        self.classmap = dict()

    def process(self) -> dict:
        pass

    def save(self, input_data):
        super(TltAnno, self).save(input_data)

        dst_dir = self.dst_dir
        image_folder = self.image_folder
        label_folder = self.label_folder
        img_extension = self.img_extension
        anno_extension = self.anno_extension
        verbose = self.verbose
        logger = self.logger

        dst_image_dir, dst_label_dir = check_input_output_folder(dst_dir, image_folder, label_folder, is_input=False)

        classmap = input_data['classmap']

        for i, (img_path, anno_obj) in enumerate(input_data['anno_data'].items(), 1):
            img_name = get_img_name(img_path)
            if verbose:
                logger.info('save {}'.format(img_name))

            size = anno_obj['size']
            objects = anno_obj['objects']

            label_list = list()
            for obj in objects:
                name = obj['name']
                bndbox = obj['bndbox']

                xmin, ymin, xmax, ymax = bndbox
                label_list.append([name, 0, 0, 0, xmin, ymin, xmax, ymax, 0, 0, 0, 0, 0, 0, 0])
            # save
            img = cv2.imread(img_path)
            dst_img_path = os.path.join(dst_image_dir, img_name + img_extension)
            cv2.imwrite(dst_img_path, img)

            dst_label_path = os.path.join(dst_label_dir, img_name + anno_extension)
            np.savetxt(dst_label_path, label_list, fmt='%s', delimiter=' ')

        if verbose:
            logger.info('save classmap.json')
        classmap_path = os.path.join(dst_dir, 'classmap.json')
        with open(classmap_path, 'w') as f:
            json.dump(classmap, f)
        if verbose:
            print(__name__ + ' done')



# -*- coding: utf-8 -*-

"""
@date: 2020/7/15 下午7:04
@file: labelimg_anno.py
@author: zj
@description: 
"""

import os
import cv2
import json
import glob
from collections import OrderedDict

from ..anno import registry
from ..anno.base_anno import BaseAnno
from ..util.utility import dict_to_xml
from ..util.misc import check_image_label, get_img_name, parse_classmap, check_input_output_folder, parse_img_path
from ..util.parse_voc_xml import ParseVocXml
from ..util.logger import setup_logger


@registry.ANNOS.register('labelimg')
class LabelImgAnno(BaseAnno):
    """
    Dataset annotated with LabelImg. By default:
    1. The image is in PNG format
    2. The annotation file is an XML file in Pascal VOC format
    3. The image and annotation file are placed in the same directory. For example 1.txt/1.xml/2.txt/2.xml/...
    As LabelImg, refer to [tzutalin/labelImg](https://github.com/tzutalin/labelImg)
    """

    def __init__(self, cfg):
        self.name = cfg.LABELIMG.NAME
        self.img_extension = cfg.LABELIMG.IMG_EXTENSION
        self.anno_extension = cfg.LABELIMG.ANNO_EXTENSION

        if cfg.ANNO.PARSER == self.name:
            self.src_dir = cfg.INPUT.DIR
            self.image_folder = cfg.INPUT.IMAGE_FOLDER
            self.label_folder = cfg.INPUT.LABEL_FOLDER
        if cfg.ANNO.CREATOR == self.name:
            self.dst_dir = cfg.OUTPUT.DIR
            self.image_folder = cfg.OUTPUT.IMAGE_FOLDER
            self.label_folder = cfg.OUTPUT.LABEL_FOLDER

        self.verbose = cfg.ANNO.VERBOSE

        self.logger = setup_logger(__name__)
        self.classmap = dict()

    def process(self) -> dict:
        src_dir = self.src_dir
        image_folder = self.image_folder
        label_folder = self.label_folder
        img_extension = self.img_extension
        anno_extension = self.anno_extension
        verbose = self.verbose
        logger = self.logger

        image_dir, label_dir = check_input_output_folder(src_dir, image_folder, label_folder, is_input=True)

        img_path_list = sorted(glob.glob(os.path.join(image_dir, '*' + img_extension)))
        anno_path_list = sorted(glob.glob(os.path.join(label_dir, '*' + anno_extension)))
        check_image_label(img_path_list, anno_path_list)

        anno_data = dict()
        for i, (img_path, anno_path) in enumerate(zip(img_path_list, anno_path_list), 1):
            if verbose:
                logger.info('parse {}'.format(anno_path))
            parser = ParseVocXml(anno_path)

            anno_obj = dict()
            anno_obj['size'] = parser.get_size()
            anno_obj['objects'] = parser.get_objects()
            anno_data[img_path] = anno_obj

            parse_classmap(self.classmap, anno_obj['objects'])

        return {'classmap': self.classmap, 'anno_data': anno_data}

    def save(self, input_data):
        super(LabelImgAnno, self).save(input_data)

        dst_dir = self.dst_dir
        image_folder = self.image_folder
        label_folder = self.label_folder
        img_extension = self.img_extension
        anno_extension = self.anno_extension
        verbose = self.verbose
        logger = self.logger

        dst_image_dir, dst_label_dir = check_input_output_folder(dst_dir, image_folder, label_folder, is_input=False)

        classmap = input_data['classmap']
        for i, (img_path, anno_obj) in enumerate(input_data['anno_data'].items(), 1):
            img_name = get_img_name(img_path)
            if verbose:
                logger.info('save {}'.format(img_name))

            root_dict = OrderedDict()
            # Root node
            annotation_dict = OrderedDict()
            root_dict['annotation'] = annotation_dict

            abs_img_path = os.path.abspath(img_path)
            folder, filename = parse_img_path(abs_img_path)[:2]
            annotation_dict['folder'] = folder
            annotation_dict['filename'] = filename
            annotation_dict['path'] = abs_img_path
            annotation_dict['source'] = OrderedDict({'database': 'Unknown'})

            size = anno_obj['size']
            if len(size) == 2:
                h, w = size[:2]
                d = 1
            else:
                h, w, d = size[:3]
            size_dict = OrderedDict()
            size_dict['width'] = w
            size_dict['height'] = h
            size_dict['depth'] = d
            annotation_dict['size'] = size_dict

            annotation_dict['segmented'] = 0

            object_list = list()
            objects = anno_obj['objects']
            for obj in objects:
                name = obj['name']
                bndbox = obj['bndbox']
                xmin, ymin, xmax, ymax = bndbox

                object_dict = OrderedDict()
                object_dict['name'] = name
                object_dict['pose'] = 'Unspecified'
                object_dict['truncated'] = 0
                object_dict['difficult'] = 0

                bndbox_dict = OrderedDict()
                bndbox_dict['xmin'] = xmin
                bndbox_dict['ymin'] = ymin
                bndbox_dict['xmax'] = xmax
                bndbox_dict['ymax'] = ymax
                object_dict['bndbox'] = bndbox_dict

                object_list.append(object_dict)
            annotation_dict['object'] = object_list
            # save
            img = cv2.imread(img_path)
            dst_img_path = os.path.join(dst_image_dir, img_name + img_extension)
            cv2.imwrite(dst_img_path, img)

            dst_label_path = os.path.join(dst_label_dir, img_name + anno_extension)
            dict_to_xml(root_dict, dst_label_path)

        if verbose:
            logger.info('save classmap.json')
        classmap_path = os.path.join(dst_dir, 'classmap.json')
        with open(classmap_path, 'w') as f:
            json.dump(classmap, f)
        if verbose:
            logger.info(__name__ + ' done')



# -*- coding: utf-8 -*-

"""
@date: 2020/7/14 下午7:39
@file: yolov5_anno.py
@author: zj
@description: 
"""

import os
import cv2
import numpy as np
import json

from . import registry
from .base_anno import BaseAnno
from ..util.utility import xyxy_2_xywh
from ..util.misc import get_img_name, check_input_output_folder
from ..util.logger import setup_logger


@registry.ANNOS.register('yolov5')
class YoLoV5Anno(BaseAnno):
    """
    Create [ultralytics/yolov5](https://github.com/ultralytics/yolov5) training data set
    The image and annotation files are arranged as follows:
    ├── yolov5-dataset
        ── images
            ├── 201.png
            ├── 202.png
        ── labels
            ├── 201.txt
            ├── 202.txt
    The format of the annotation file is as follows:
    class_num x_center y_center width height
    class_num is the category, starting from 0
    x_center/y_center/width/height use coordinates relative to the width and height of the image
    Examples are as follows:
    0 0.110547 0.051621 0.154554 0.023163

    Save one extra classmap.txt , save class_num and the corresponding class name
    """

    def __init__(self, cfg):
        self.name = cfg.YOLOV5.NAME
        self.img_extension = cfg.YOLOV5.IMG_EXTENSION
        self.anno_extension = cfg.YOLOV5.ANNO_EXTENSION

        if cfg.ANNO.PARSER == self.name:
            self.src_dir = cfg.INPUT.DIR
            self.image_folder = cfg.INPUT.IMAGE_FOLDER
            self.label_folder = cfg.INPUT.LABEL_FOLDER
        if cfg.ANNO.CREATOR == self.name:
            self.dst_dir = cfg.OUTPUT.DIR
            self.image_folder = cfg.OUTPUT.IMAGE_FOLDER
            self.label_folder = cfg.OUTPUT.LABEL_FOLDER

        self.verbose = cfg.ANNO.VERBOSE

        self.logger = setup_logger(__name__)
        self.classmap = dict()

    def process(self) -> dict:
        pass

    def save(self, input_data):
        super(YoLoV5Anno, self).save(input_data)

        dst_dir = self.dst_dir
        image_folder = self.image_folder
        label_folder = self.label_folder
        img_extension = self.img_extension
        anno_extension = self.anno_extension
        verbose = self.verbose
        logger = self.logger

        dst_image_dir, dst_label_dir = check_input_output_folder(dst_dir, image_folder, label_folder, is_input=False)

        classmap = input_data['classmap']
        for i, (img_path, anno_obj) in enumerate(input_data['anno_data'].items(), 1):
            img_name = get_img_name(img_path)
            if verbose:
                logger.info('save {}'.format(img_name))

            size = anno_obj['size']
            objects = anno_obj['objects']

            label_list = list()
            for obj in objects:
                name = obj['name']
                bndbox = obj['bndbox']

                x_center, y_center, box_w, box_h = xyxy_2_xywh(bndbox, size)
                label_list.append([classmap[name], x_center, y_center, box_w, box_h])
            # 保存
            img = cv2.imread(img_path)
            dst_img_path = os.path.join(dst_image_dir, img_name + img_extension)
            cv2.imwrite(dst_img_path, img)

            dst_label_path = os.path.join(dst_label_dir, img_name + anno_extension)
            np.savetxt(dst_label_path, label_list, fmt='%d %.6f %.6f %.6f %.6f', delimiter=' ')

        if verbose:
            logger.info('save classmap.json')
        classmap_path = os.path.join(dst_dir, 'classmap.json')
        with open(classmap_path, 'w') as f:
            json.dump(classmap, f)
        if verbose:
            logger.info(__name__ + ' done')



# -*- coding: utf-8 -*-

"""
@date: 2021/4/3 下午3:26
@file: lmdb.py
@author: zj
@description: 
"""

import os
import lmdb
import pickle
from tqdm import tqdm

from . import registry
from .base_anno import BaseAnno
from ..util.logger import setup_logger


def dumps_data(obj):
    """
    Serialize an object.
    Returns:
        Implementation-dependent bytes-like object
    """
    return pickle.dumps(obj)


@registry.ANNOS.register('lmdb')
class LMDB(BaseAnno):

    def __init__(self, cfg) -> None:
        self.name = cfg.LMDB.NAME

        if cfg.ANNO.CREATOR == self.name:
            self.dst_dir = cfg.OUTPUT.DIR
            self.image_folder = cfg.OUTPUT.IMAGE_FOLDER

        self.verbose = cfg.ANNO.VERBOSE
        self.logger = setup_logger(__name__)

    def process(self) -> dict:
        pass

    def save(self, input_data: dict):
        verbose = self.verbose
        logger = self.logger

        lmdb_path = os.path.join(self.dst_dir, f'{self.image_folder}.lmdb')
        if verbose:
            logger.info(f"Generate LMDB to {lmdb_path}")

        data_loader = input_data['dataloader']
        classes = input_data['classes']
        self.folder2lmdb(lmdb_path, data_loader, classes)

        if self.verbose:
            logger.info(__name__ + ' done')

    def folder2lmdb(self, dpath, data_loader, classes, write_frequency=5000):
        verbose = self.verbose
        logger = self.logger

        db = lmdb.open(dpath, subdir=False,
                       map_size=1099511627776 * 2, readonly=False,
                       meminit=False, map_async=True)

        txn = db.begin(write=True)
        idx = 0
        for image, label in tqdm(data_loader):
            txn.put(u'{}'.format(idx).encode('ascii'), dumps_data((image, label)))

            idx += 1
            if idx % write_frequency == 0:
                # if verbose:
                #     logger.info("[%d/%d]" % (idx, len(data_loader)))
                txn.commit()
                txn = db.begin(write=True)

        # finish iterating through dataset
        txn.commit()
        keys = [u'{}'.format(k).encode('ascii') for k in range(len(data_loader))]
        with db.begin(write=True) as txn:
            txn.put(b'__keys__', dumps_data(keys))
            txn.put(b'__len__', dumps_data(len(keys)))
            txn.put(b'classes', dumps_data(classes))

        if verbose:
            logger.info("Flushing database ...")
        db.sync()
        db.close()
